// src/from-type.ts
import {
  any,
  array,
  bigint,
  blob,
  boolean,
  custom,
  date,
  definitionReference,
  file,
  formData,
  func,
  intersection,
  iterator,
  literal,
  map,
  named,
  never,
  nullType,
  number,
  object,
  optional,
  record,
  regexp,
  set,
  string,
  symbol as symbolType,
  tuple,
  undefinedType,
  union,
  unionIfMultiple,
  unknown,
  voidType
} from "@onetyped/core";
import ts from "typescript";
var hasFlag = (type, flag) => {
  return (type.flags & flag) === flag;
};
var hasSymbolFlag = (symbol, flag) => {
  return (symbol.flags & flag) === flag;
};
var hasObjectFlag = (type, flag) => {
  return (type.objectFlags & flag) === flag;
};
var getStringHash = (string_) => {
  let hash = 0;
  for (let index = 0, length = string_.length; index < length; index++) {
    const chr = string_.charCodeAt(index);
    hash = (hash << 5) - hash + chr;
    hash = Math.trunc(hash);
  }
  return hash >>> 0;
};
var getNodeFromCallSignatures = (callSignatures, options) => {
  if (callSignatures.length > 0) {
    const nodeCallSignatures = callSignatures.map((signature) => {
      const parameters = signature.getParameters().map((parameter) => {
        const parameterType = options.checker.getTypeOfSymbolAtLocation(
          parameter,
          options.locationNode
        );
        return fromTypeInternal({
          ...options,
          type: parameterType
        });
      });
      const returnType = options.checker.getReturnTypeOfSignature(signature);
      const functionNode = func({
        arguments: parameters,
        return: fromTypeInternal({
          ...options,
          type: returnType
        })
      });
      return functionNode;
    });
    return unionIfMultiple(nodeCallSignatures);
  }
};
var fromType = (options) => {
  const typeCache = options.typeCache ?? /* @__PURE__ */ new Map();
  const customTypes = options.customTypes ?? /* @__PURE__ */ new Map();
  const definitionNodeType = options.definitionNodeType ?? "named";
  return {
    node: fromTypeInternal({
      ...options,
      typeCache,
      customTypes,
      definitionNodeType
    }),
    typeCache
  };
};
var createOrReferenceSymbolDefinition = (typeArguments, options) => {
  const { type, checker, typeCache, definitionNodeType } = options;
  const idObjectHash = JSON.stringify({
    id: type.id,
    typeArgIds: typeArguments.map(
      (typeArgument) => typeArgument.id
    )
  });
  const hash = `type_${getStringHash(idObjectHash)}`;
  const typeString = checker.typeToString(type);
  const symbolNode = typeCache.get(hash);
  if (symbolNode) {
    if (definitionNodeType === "definitionReference") {
      return definitionReference(hash, typeString);
    }
    return symbolNode;
  }
  const newType = Object.assign(
    Object.create(Object.getPrototypeOf(type)),
    type
  );
  newType._mustResolve = true;
  if (definitionNodeType === "definitionReference") {
    typeCache.set(hash, unknown());
    const node2 = fromTypeInternal({
      ...options,
      type: newType
    });
    typeCache.set(hash, node2);
    return definitionReference(hash, typeString);
  }
  const node = named(null, typeString);
  typeCache.set(hash, node);
  node.node = fromTypeInternal({
    ...options,
    type: newType
  });
  return node.node;
};
var fromTypeInternal = (options) => {
  const { type, checker, locationNode, customTypes } = options;
  if (hasFlag(type, ts.TypeFlags.String)) {
    return string();
  }
  if (hasFlag(type, ts.TypeFlags.Number)) {
    return number();
  }
  if (hasFlag(type, ts.TypeFlags.Boolean)) {
    return boolean();
  }
  if (hasFlag(type, ts.TypeFlags.Null)) {
    return nullType();
  }
  if (hasFlag(type, ts.TypeFlags.Undefined)) {
    return undefinedType();
  }
  if (hasFlag(type, ts.TypeFlags.Unknown)) {
    return unknown();
  }
  if (type.isLiteral()) {
    if (typeof type.value === "object") {
      const bigIntLiteral = BigInt(
        `${type.value.negative ? "-" : ""}${type.value.base10Value}`
      );
      return literal(bigIntLiteral);
    }
    return literal(type.value);
  }
  if (hasFlag(type, ts.TypeFlags.BooleanLiteral)) {
    const literalType = checker.typeToString(type);
    return literal(literalType === "true");
  }
  if (hasFlag(type, ts.TypeFlags.TemplateLiteral)) {
    return literal(checker.typeToString(type));
  }
  if (hasFlag(type, ts.TypeFlags.Void)) {
    return voidType();
  }
  if (hasFlag(type, ts.TypeFlags.Any)) {
    return any();
  }
  if (hasFlag(type, ts.TypeFlags.Never)) {
    return never();
  }
  if (hasFlag(type, ts.TypeFlags.BigInt)) {
    return bigint();
  }
  const symbol = type.getSymbol();
  if (symbol) {
    const customType = customTypes.get(symbol.id);
    if (customType) {
      const arguments_ = type.aliasTypeArguments?.map((argument) => {
        return fromTypeInternal({
          ...options,
          type: argument
        });
      });
      return custom(customType.name, arguments_);
    }
    switch (symbol.name) {
      case "Array":
      case "ReadonlyArray": {
        const typeArguments = checker.getTypeArguments(type);
        return array(
          fromTypeInternal({
            ...options,
            type: typeArguments[0]
          })
        );
      }
      case "Iterable":
      case "IterableIterator": {
        const typeArguments = checker.getTypeArguments(type);
        return iterator(
          fromTypeInternal({
            ...options,
            type: typeArguments[0]
          })
        );
      }
      case "Date": {
        return date();
      }
      case "RegExp": {
        return regexp();
      }
      case "Set": {
        const typeArguments = checker.getTypeArguments(type);
        return set(fromTypeInternal({ ...options, type: typeArguments[0] }));
      }
      case "Map": {
        const typeArguments = checker.getTypeArguments(type);
        return map(
          fromTypeInternal({
            ...options,
            type: typeArguments[0]
          }),
          fromTypeInternal({
            ...options,
            type: typeArguments[1]
          })
        );
      }
      case "FormData": {
        return formData();
      }
      case "File": {
        return file();
      }
      case "Blob": {
        return blob();
      }
    }
    if (symbol.name !== "__type" && symbol.name !== "__object" && !type._mustResolve) {
      return createOrReferenceSymbolDefinition(
        checker.getTypeArguments(type),
        options
      );
    }
  }
  const aliasSymbol = type.aliasSymbol;
  const aliasTypeArguments = type.aliasTypeArguments;
  if (aliasSymbol) {
    const customType = customTypes.get(aliasSymbol.id);
    if (customType) {
      const arguments_ = aliasTypeArguments?.map((argument) => {
        return fromTypeInternal({
          ...options,
          type: argument
        });
      });
      return custom(customType.name, arguments_);
    }
  }
  if (aliasSymbol?.escapedName === "Record" && aliasTypeArguments?.length === 2) {
    return record(
      fromTypeInternal({
        ...options,
        type: aliasTypeArguments[0]
      }),
      fromTypeInternal({
        ...options,
        type: aliasTypeArguments[1]
      })
    );
  }
  if (aliasSymbol && !type._mustResolve) {
    return createOrReferenceSymbolDefinition(aliasTypeArguments ?? [], options);
  }
  if (hasFlag(type, ts.TypeFlags.Object)) {
    const objectType = type;
    if (hasObjectFlag(objectType, ts.ObjectFlags.Reference)) {
      const { target } = type;
      if (hasObjectFlag(target, ts.ObjectFlags.Tuple)) {
        const tupleType = type;
        const target2 = tupleType.target;
        const tupleTypes = tupleType.typeArguments?.map((typeArgument, index) => {
          const elementFlags = target2.elementFlags[index];
          const node = fromTypeInternal({
            ...options,
            type: typeArgument
          });
          if (elementFlags & ts.ElementFlags.Optional) {
            return optional(node);
          }
          return node;
        }) ?? [];
        return tuple(tupleTypes);
      }
    }
    const nodes = [];
    const keyTypes = [];
    let valueType;
    const stringIndexType = objectType.getStringIndexType();
    if (stringIndexType) {
      keyTypes.push(string());
      valueType = fromTypeInternal({
        ...options,
        type: stringIndexType
      });
    }
    const numberIndexType = objectType.getNumberIndexType();
    if (numberIndexType) {
      keyTypes.push(number());
      if (!valueType) {
        valueType = fromTypeInternal({
          ...options,
          type: numberIndexType
        });
      }
    }
    if (keyTypes.length > 0 && valueType) {
      nodes.push(
        record(
          unionIfMultiple(
            keyTypes
          ),
          valueType
        )
      );
    }
    const propertyEntries = type.getProperties().map((property) => {
      const propertyType = checker.getTypeOfSymbolAtLocation(
        property,
        locationNode
      );
      let node = fromTypeInternal({
        ...options,
        type: propertyType
      });
      if (hasSymbolFlag(property, ts.SymbolFlags.Optional)) {
        node = optional(node);
      }
      const isUniqueSymbol = (() => {
        if (hasFlag(propertyType, ts.TypeFlags.UniqueESSymbol)) {
          return true;
        }
        const propertyWithLink = property;
        if (propertyWithLink.links?.nameType && hasFlag(propertyWithLink.links.nameType, ts.TypeFlags.UniqueESSymbol)) {
          return true;
        }
        return false;
      })();
      const symbolType2 = /^__@([^@]+)@/.exec(property.name)?.[1];
      if (isUniqueSymbol && symbolType2) {
        switch (symbolType2) {
          case "toStringTag":
          case "iterator":
          case "asyncIterator":
          case "unscopables": {
            return [`[Symbol.${symbolType2}]`, node];
          }
          default: {
            return [`__uniqueSymbol__${symbolType2}__`, node];
          }
        }
      }
      return [property.name, node];
    });
    const callSignatures = type.getCallSignatures();
    const functionNode = getNodeFromCallSignatures(callSignatures, options);
    if (functionNode) {
      nodes.push(functionNode);
    }
    if (propertyEntries.length > 0) {
      const propertySchemas = Object.fromEntries(propertyEntries);
      nodes.push(object(propertySchemas));
    }
    if (nodes.length === 1) {
      return nodes[0];
    }
    return intersection(nodes);
  }
  if (type.isUnion()) {
    const unionTypes = type.types.map(
      (unionType) => fromTypeInternal({
        ...options,
        type: unionType
      })
    );
    return union(unionTypes);
  }
  if (type.isIntersection()) {
    const intersectionTypes = type.types.map(
      (type2) => fromTypeInternal({
        ...options,
        type: type2
      })
    );
    return intersection(intersectionTypes);
  }
  if (hasFlag(type, ts.TypeFlags.UniqueESSymbol) || "intrinsicName" in type && type.intrinsicName === "symbol") {
    return symbolType(checker.typeToString(type));
  }
  console.debug("unhandled type", checker.typeToString(type));
  return symbolType("TRPC_DRIFT_UNHANDLED_TYPE");
};

// src/to-type-node.ts
import { isNodeOptional, mapMultipleInnerNode } from "@onetyped/core";
import ts2 from "typescript";
var { factory: f } = ts2;
var printNode = (node, printerOptions) => {
  const sourceFile = ts2.createSourceFile(
    "print.ts",
    "",
    ts2.ScriptTarget.Latest,
    false,
    ts2.ScriptKind.TS
  );
  const printer = ts2.createPrinter(printerOptions);
  return printer.printNode(ts2.EmitHint.Unspecified, node, sourceFile);
};
var toTypeNode = (node) => {
  switch (node.typeName) {
    case "definitionReference": {
      return f.createTypeReferenceNode(node.text);
    }
    case "string": {
      return f.createKeywordTypeNode(ts2.SyntaxKind.StringKeyword);
    }
    case "number": {
      return f.createKeywordTypeNode(ts2.SyntaxKind.NumberKeyword);
    }
    case "boolean": {
      return f.createKeywordTypeNode(ts2.SyntaxKind.BooleanKeyword);
    }
    case "null": {
      return f.createLiteralTypeNode(f.createNull());
    }
    case "undefined": {
      return f.createKeywordTypeNode(ts2.SyntaxKind.UndefinedKeyword);
    }
    case "symbol": {
      return f.createKeywordTypeNode(ts2.SyntaxKind.SymbolKeyword);
    }
    case "literal": {
      if (node.type === void 0) {
        return f.createKeywordTypeNode(ts2.SyntaxKind.UndefinedKeyword);
      }
      switch (typeof node.type) {
        case "string": {
          return f.createLiteralTypeNode(f.createStringLiteral(node.type));
        }
        case "number": {
          return f.createLiteralTypeNode(f.createNumericLiteral(node.type));
        }
        case "boolean": {
          return f.createLiteralTypeNode(
            node.type === true ? f.createTrue() : f.createFalse()
          );
        }
        case "bigint": {
          return f.createLiteralTypeNode(
            f.createBigIntLiteral(node.type.toString())
          );
        }
      }
    }
    case "unknown": {
      return f.createKeywordTypeNode(ts2.SyntaxKind.UnknownKeyword);
    }
    case "void": {
      return f.createKeywordTypeNode(ts2.SyntaxKind.VoidKeyword);
    }
    case "any": {
      return f.createKeywordTypeNode(ts2.SyntaxKind.AnyKeyword);
    }
    case "never": {
      return f.createKeywordTypeNode(ts2.SyntaxKind.NeverKeyword);
    }
    case "regexp": {
      return f.createTypeReferenceNode("RegExp", []);
    }
    case "bigint": {
      return f.createKeywordTypeNode(ts2.SyntaxKind.BigIntKeyword);
    }
    case "date": {
      return f.createTypeReferenceNode("Date", []);
    }
    case "file": {
      return f.createTypeReferenceNode("File", []);
    }
    case "blob": {
      return f.createTypeReferenceNode("Blob", []);
    }
    case "formData": {
      return f.createTypeReferenceNode("FormData", []);
    }
    case "object": {
      return f.createTypeLiteralNode(
        Object.entries(node.shape).map(([key, value]) => {
          return f.createPropertySignature(
            void 0,
            key,
            isNodeOptional(value) ? f.createToken(ts2.SyntaxKind.QuestionToken) : void 0,
            toTypeNode(value)
          );
        })
      );
    }
    case "array": {
      return f.createArrayTypeNode(toTypeNode(node.type));
    }
    case "set": {
      return f.createTypeReferenceNode("Set", [toTypeNode(node.type)]);
    }
    case "iterator": {
      return f.createTypeReferenceNode("Iterable", [toTypeNode(node.type)]);
    }
    case "record": {
      return f.createTypeReferenceNode("Record", [
        toTypeNode(node.key),
        toTypeNode(node.value)
      ]);
    }
    case "map": {
      return f.createTypeReferenceNode("Map", [
        toTypeNode(node.key),
        toTypeNode(node.value)
      ]);
    }
    case "function": {
      return f.createFunctionTypeNode(
        void 0,
        node.arguments.map(
          (argument, index) => f.createParameterDeclaration(
            void 0,
            void 0,
            `args_${index}`,
            void 0,
            toTypeNode(argument)
          )
        ),
        node.return ? toTypeNode(node.return) : f.createKeywordTypeNode(ts2.SyntaxKind.VoidKeyword)
      );
    }
    case "custom": {
      return f.createTypeReferenceNode(
        node.name,
        node.arguments?.map((argument) => toTypeNode(argument))
      );
    }
    case "union": {
      return f.createUnionTypeNode(mapMultipleInnerNode(node, toTypeNode));
    }
    case "tuple": {
      return f.createTupleTypeNode(mapMultipleInnerNode(node, toTypeNode));
    }
    case "intersection": {
      return f.createIntersectionTypeNode(
        mapMultipleInnerNode(node, toTypeNode)
      );
    }
    case "named": {
      return f.createTypeReferenceNode(node.name);
    }
    case "circular": {
      return f.createLiteralTypeNode(f.createStringLiteral("[Circular]"));
    }
  }
};
export {
  fromType,
  printNode,
  toTypeNode
};
